import websocket
import threading
import time
import json

class TwitchWebSocket:
  WEBSOCKET_URL = "wss://eventsub-beta.wss.twitch.tv/ws"
  SESSION_ID = ""
  
  MSG_CALLBACKS = { }
  CLOSE_CB = lambda wsapp, close_status_code, close_msg: None 
  
  def __init__(self, url_override = None):
    self.MSG_CALLBACKS['default']           = self.default_message_handler
    self.MSG_CALLBACKS['session_welcome']   = self.handle_welcome_message
    self.MSG_CALLBACKS['session_keepalive'] = self.handle_keepalive_message
    self.MSG_CALLBACKS['session_reconnect'] = self.handle_reconnect_message
    self.MSG_CALLBACKS['revocation']        = self.handle_revocation_message
    
    if url_override:
      self.WEBSOCKET_URL = url_override
    
    self.connect()
    
  def add_callback(self, message_type, callback):
    self.MSG_CALLBACKS[message_type] = callback
    
  def connect(self, wait_for_welcome = True):
    self.CONNECTED = False
    self.SESSION_ID = ""
    self.WS = websocket.WebSocketApp(self.WEBSOCKET_URL, on_message = self.on_message, on_close = self.on_close)
    self.THREAD = threading.Thread(target = self.WS.run_forever, daemon = True)
    self.THREAD.start()
    
    time_waiting = 0
    while time_waiting < 10.0 and wait_for_welcome and self.SESSION_ID == "":
      time.sleep(0.1)
      time_waiting += 0.1
      
    if not self.CONNECTED:
      self.WS.close()
      print("Server did not send welcome message. Connection failed.")
    
  def disconnect(self):
    self.WS.close()
    
  def handle_welcome_message(self, wsapp, message):
    print('Handling welcome message.')
    msg_json = json.loads(message)
    try:
      self.SESSION_ID = msg_json['payload']['session']['id']
      self.CONNECTED = True
    except:
      print("Invalid welcome message format.")
      self.CONNECTED = False
      
  def handle_reconnect_message(self, wsapp, message):
    msg_json = json.loads(message)
    
    self.CONNECTED = False
    self.SESSION_ID = ""
    
    new_ws = websocket.WebSocketApp(url = msg_json['payload']['session']['reconnect_url'], on_message = self.on_message, on_close = self.on_close)
    new_thread = threading.Thread(target = new_ws.run_forever, daemon = True)
    new_thread.start()
    
    time_waiting = 0
    while time_waiting < 10.0 and not self.CONNECTED:
      time.sleep(0.1)
      time_waiting += 0.1
      
    if self.CONNECTED:
      self.WS.close()
      self.WS = new_ws
      self.THREAD = new_thread
    else:
      print("Failed to reconnect. Connection lost.")
      new_ws.close()
      self.WS.close()
      
  def handle_revocation_message(self, wsapp, message):
    msg_json = json.loads(message)
    
    status = msg_json['payload']['subscription']['status']
    
    if status == "user_removed":
      print("")
    
  def handle_keepalive_message(self, wsapp, message):
    None
    
  def default_message_handler(self, wsapp, message):
    print("Unhandled message received.")
    print(f"Content: {message}")
    
  def on_message(self, wsapp, message):
    msg_json = json.loads(message)
    message_type = msg_json['metadata']['message_type']
    
    if message_type == 'notification':
      notif_type = msg_json['payload']['subscription']['type']
      if notif_type in self.MSG_CALLBACKS:
        self.MSG_CALLBACKS[notif_type](wsapp, message)
      else:
        self.MSG_CALLBACKS['default'](wsapp, message)
    else:
      if message_type in self.MSG_CALLBACKS:
        self.MSG_CALLBACKS[message_type](wsapp, message)
      else:
        self.MSG_CALLBACKS['default'](wsapp, message)
    
  def on_close(self, wsapp, close_status_code, close_msg):
    print(f'{close_msg} (Status: {close_status_code})')
    self.SESSION_ID = ""
    self.CONNECTED  = False


from .saferequests import *
from .websocket import *
from .events import *

from typing import Callable, Tuple

import time

TWITCH_API_TIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

class TwitchAPI:
  API_URL = "https://api.twitch.tv/helix"
  CLIENT_ID = ""
  CLIENT_SECRET = ""
  ACCESS_TOKEN = ""
  DEFAULT_HEADERS = {}
  rlrequests = RateLimitedRequests(400, 60)
  TWITCH_WEBSOCKET = None
  
  def __add_parameters(self, url : str, params : dict):
    """Add parameters to an API request URL.

    Args:
        url (string): API endpoint
        params (string): Dictionary of parameters. See function descriptions for valid params.

    Returns:
        string: URL with params.
    """
    separator = "?"
    for k, v in params.items():
      url += f"{separator}{k}={v}"
      separator = "&"
    return url
  
  def __raise_req_error(self, response : requests.Response):
    raise Exception(f'Status {response.status_code}: {response.content}')

  def __init__(self, credentials : dict):
    """Constructor for TwitchAPI. Must pass in credentials in the form of a dictionary.

    Args:
        credentials (dict): API Credentials. "CLIENT_ID" and "CLIENT_SECRET" should be in the dict.
    """
    self.CLIENT_ID = credentials["CLIENT_ID"]
    
    if 'ACCESS_TOKEN' in credentials:
      self.ACCESS_TOKEN = credentials["ACCESS_TOKEN"]
    else:
      self.CLIENT_SECRET = credentials["CLIENT_SECRET"]
    
      r = requests.post(f'https://id.twitch.tv/oauth2/token?client_id={self.CLIENT_ID}&client_secret={self.CLIENT_SECRET}&grant_type=client_credentials', headers = {'Content-Type': 'application/x-www-form-urlencoded'})
    
      try:
        self.ACCESS_TOKEN = r.json()['access_token']
      except:
        raise Exception("Failed to create access token. Invalid credentials.")
    
    self.DEFAULT_HEADERS = { "Authorization": f"Bearer {self.ACCESS_TOKEN}", "Client-Id": self.CLIENT_ID }
    
  def refresh_access_token(self, refresh_token : str) -> dict:
    url = f'https://id.twitch.tv/oauth2/token'
    post_params = {
      'client_id': self.CLIENT_ID,
      'client_secret': self.CLIENT_SECRET,
      'grant_type': 'refresh_token',
      'refresh_token': refresh_token
    }
    url = self.__add_parameters(url, post_params)
    
    r : requests.Response = self.rlrequests.post(url = url, headers = { 'Content-Type': 'application/x-www-form-urlencoded' })
    
    try:
      self.ACCESS_TOKEN = r.json()['access_token']
    except:
      raise Exception(f"{r.status_code}: {r.content}.\nFailed to refresh access token.")
    
    self.DEFAULT_HEADERS = { "Authorization": f"Bearer {self.ACCESS_TOKEN}", "Client-Id": self.CLIENT_ID }
    
    return r

  def get_user_id(self, login : str = "") -> str:
    """Get user ID from username.

    Args:
        username (string): Username

    Returns:
        string: User ID
    """
    if login == "":
      url = f'{self.API_URL}/users'
    else:
      url = f"{self.API_URL}/users?login={login}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()["data"][0]["id"]
    else:
      self.__raise_req_error(r)
  
  def get_channel_info(self, user_id : str) -> dict:
    """Get Channel Information.

    Args:
        user_id (string): User ID

    Returns:
        dict: Channel information
    """
    url = f"{self.API_URL}/channels?broadcaster_id={user_id}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data'][0]
    else:
      self.__raise_req_error(r)
      
  def get_category_info(self, category_name : str) -> dict:
    """Get category information from category name

    Args:
        category_name (string): Category name

    Returns:
        dict: Category information
    """
    url = f"{self.API_URL}/games?name={category_name}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()["data"][0]
    else:
      self.__raise_req_error(r)
      
  def get_category_id(self, category_name : str) -> str:
    """Get category ID

    Args:
        category_name (str): _description_

    Returns:
        str: _description_
    """
    return self.get_category_info(category_name)['id']

  def get_clip(self, clip_id : str) -> dict:
    """Get info for one clip from ID.

    Args:
        clip_id (string): Video ID

    Returns:
        dict: clip info
    """
    url = f"{self.API_URL}/clips?id={clip_id}"
    r: requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data'][0]
    else:
      self.__raise_req_error(r)

  def get_clips(self, params : dict) -> Tuple[list, str]:
    """Get clips based on params.

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            id (string): Clip ID
            broadcaster_id (string): Broadcaster ID
            game_id (string): Game/Category ID
            started_at (string): RFC3339 format, use TWITCH_API_TIME_FORMAT from this library
            ended_at (string): RFC3339 format, use TWITCH_API_TIME_FORMAT from this library
            first (int): fetch the first n clips
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of clip info
    """
    url = f"{self.API_URL}/clips"
    url = self.__add_parameters(url, params)
    
    r : requests.Response = self.rlrequests.get(url = url, headers=self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      try:
        return r.json()['data'], r['pagination']['cursor']
      except:
        return r.json()['data'], ""
    else:
      self.__raise_req_error(r)

  def get_all_clips(self, params : dict) -> list:
    """Get all clips based on params (auto-pagination).

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            id (string): Clip ID
            broadcaster_id (string): Broadcaster ID
            game_id (string): Game/Category ID
            started_at (string): RFC3339 format, use TWITCH_API_TIME_FORMAT from this library
            ended_at (string): RFC3339 format, use TWITCH_API_TIME_FORMAT from this library
            first (int): fetch the first n clips
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of clip info
    """
    all_clips = []
    while True:
      clips, cursor = self.get_clips(params)

      for clip in clips:
        all_clips.append(clip)
      
      if cursor == "":
        return all_clips
      else:
        params["after"] = cursor

  def get_video(self, video_id : str) -> dict:
    """Get info for one video from ID.

    Args:
        video_id (string): Video ID

    Returns:
        dict: Video info
    """
    url = f"{self.API_URL}/videos?id={video_id}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data'][0]
    else:
      self.__raise_req_error(r)

  def get_videos(self, params : dict) -> Tuple[list, str]:
    """Get videos based on params.

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            id (string): Video ID
            user_id (string): User ID
            game_id (string): Game/Category ID
            language (string): ISO 639-1
            period (string): "all", "day", "month", or "week"
            sort (string): "time", "trending", or "views"
            type (string): "all", "archive", "highlight", or "upload"
            first (int): fetch the first n videos
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of video info
        string: Pagination cursor
    """
    url = f"{self.API_URL}/videos"
    url = self.__add_parameters(url, params)
    
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      try:
        return r.json()['data'], r['pagination']['cursor']
      except:
        return r.json()['data'], ""
    else:
      self.__raise_req_error(r)

  def get_all_videos(self, params : dict) -> list:
    """Get all videos based on params (auto-pagination).

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            id (string): Video ID
            user_id (string): User ID
            game_id (string): Game/Category ID
            language (string): ISO 639-1
            period (string): "all", "day", "month", or "week"
            sort (string): "time", "trending", or "views"
            type (string): "all", "archive", "highlight", or "upload"
            first (int): fetch the first n videos
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of video info
    """
    all_clips = []
    while True:
      vids, cursor = self.get_videos(params)

      for vod in vids:
        all_clips.append(vod)
      
      if cursor == "":
        return all_clips
      else:
        params["after"] = cursor

  def get_streams(self, params : dict) -> dict:
    """Get a list of streams.

    Args:
        params (dict): Dictionary of parameters for the API request. The valid params are:
            user_id (string): User ID
            user_login (string): Username
            game_id (string): Game/Category ID
            type (string): "all" or "live"
            language (string): ISO 639-1
            first (int): fetch the first n streams
            before (string): reverse pagination
            after (string): forward pagination

    Returns:
        list: list of stream info
    """
    url = f"{self.API_URL}/streams/"
    url = self.__add_parameters(url, params)
    
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data']
    else:
      self.__raise_req_error(r)

  def is_user_live(self, user_id : str) -> bool:
    stream_info = self.get_streams({ "user_id": user_id })
    return (len(stream_info) > 0)
  
  def get_emotes(self, user_id : str) -> list:
    """Get Channel Emotes

    Args:
        user_id (string): User ID

    Returns:
        list: List of emote information.
    """
    url = f"{self.API_URL}/chat/emotes?broadcaster_id={user_id}"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data']
    else:
      self.__raise_req_error(r)
  
  def get_global_emotes(self) -> list:
    """Get Global Emotes.

    Returns:
        list: List of global emote information.
    """
    url = f"{self.API_URL}/chat/emotes/global"
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data']
    else:
      self.__raise_req_error(r)
    
  def get_custom_rewards(self, params : dict) -> list:
    """Get a list of custom rewards for a channel.

    Args:
        params (dict): dictionary of parameters for the request
          broadcaster_id (int): The ID of the broadcaster whose custom rewards you want to get
          id (string): Rewards ID
          only_manageable_rewards (bool): only get custom rewards that the app may manage

    Returns:
        _type_: _description_
    """
    url = f"{self.API_URL}/channel_points/custom_rewards"
    url = self.__add_parameters(url, params)
    
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data']
    else:
      self.__raise_req_error(r)
    
  def create_custom_reward(self, broadcaster_id : str, reward_params : dict) -> dict:
    url = f"{self.API_URL}/channel_points/custom_rewards?broadcaster_id={broadcaster_id}"
    
    r : requests.Response = self.rlrequests.post(url = url, headers = self.DEFAULT_HEADERS, json = reward_params)
    
    if r.status_code == 200:
      return r.json()['data'][0]
    else:
      self.__raise_req_error(r)
    
  def setup_websocket(self, url_override = None):
    """Connect to the Twitch WebSockets interface for subscribing to events.

    Args:
        url_override (str, optional): Pass a different websockets url for testing purposes. Defaults to None.
    """
    self.TWITCH_WEBSOCKET = TwitchWebSocket(url_override)
    
  def websocket_session_id(self) -> str:
    return self.TWITCH_WEBSOCKET.SESSION_ID
    
  def join_websocket_thread(self):
    self.TWITCH_WEBSOCKET.THREAD.join()
    
  def _add_subscription(self, params : dict) -> bool:
    if not self.TWITCH_WEBSOCKET.CONNECTED:
      print("WebSocket is not connected.")
      return False
    
    url = f'{self.API_URL}/eventsub/subscriptions'
    
    r : requests.Response = self.rlrequests.post(url = url, headers = self.DEFAULT_HEADERS, json=params)
    
    if r.status_code == 202:
      return r.json()['data'][0]['status'] == 'enabled'
    else:
      return False
    
  def get_active_subscriptions(self) -> list:
    """Get active subscriptions in current WebSocket instance.

    Returns:
        list: list of active subscriptions
    """
    url = f'{self.API_URL}/eventsub/subscriptions'
    r : requests.Response = self.rlrequests.get(url = url, headers = self.DEFAULT_HEADERS)
    
    if r.status_code == 200:
      return r.json()['data']
    else:
      self.__raise_req_error(r)
    
  def add_subscription(self, event : TwitchEvent, callback) -> bool:
    """Add a subscription to the WebSocket interface.

    Args:
        event (TwitchEvent): Desired subscription event type
        callback (function): Callback for handling notifications matching this subscription

    Returns:
        bool: Success.
    """
    self.TWITCH_WEBSOCKET.add_callback(event.notification_type(), callback)
    return self._add_subscription(event.params())
  
  def subscribe_to_updates(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(UpdateEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_follows(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(FollowEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_subscriptions(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(SubscribeEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_gifted_subscriptions(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(SubscriptionGiftEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_subscription_messages(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(SubscriptionMessageEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_cheers(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(CheerEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_raids(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(RaidEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_bans(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(BanEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_unbans(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(UnbanEvent(user_id, self.websocket_session_id()), callback)
  
  def subscribe_to_reward_redemption(self, user_id : str, callback : Callable[[object, str], None], reward_id : str = None) -> bool:
    return self.add_subscription(CustomRewardRedemptionAddEvent(user_id, self.websocket_session_id(), reward_id), callback)
    
  def subscribe_to_stream_online(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(StreamOnlineEvent(user_id, self.websocket_session_id()), callback)
    
  def subscribe_to_stream_offline(self, user_id : str, callback : Callable[[object, str], None]) -> bool:
    return self.add_subscription(StreamOfflineEvent(user_id, self.websocket_session_id()), callback)


class TwitchEvent:
  _TYPE = ''
  _VERSION = 1
  _CONDITION = {}
  _TRANSPORT = {}
  def __init__(self, type, version, condition, transport):
    self._TYPE = type
    self._VERSION = version
    self._CONDITION = condition
    self._TRANSPORT = transport
    
  def notification_type(self):
    return self._TYPE
    
  def params(self):
    return {
      'type': self._TYPE,
      'version': self._VERSION,
      'condition': self._CONDITION,
      'transport': self._TRANSPORT
    }
  
class UpdateEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.update', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class FollowEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.follow', 
                     version =  2, 
                     condition =  { 'broadcaster_user_id': user_id, 'moderator_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class SubscribeEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.subscribe', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class SubscriptionEndEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.subscription.end', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class SubscriptionGiftEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.subscription.gift', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class SubscriptionMessageEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.subscription.message', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class CheerEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.cheer', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class RaidEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.raid', 
                     version =  1, 
                     condition =  { 'to_broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class BanEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.ban', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class UnbanEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.unban', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class ModeratorAddEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.moderator.add', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class ModeratorRemoveEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.moderator.remove', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class CustomRewardAddEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'channel.channel_points_custom_reward.add', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class CustomRewardUpdateEvent(TwitchEvent):
  def __init__(self, user_id, session_id, reward_id = None):
    condition = { 'broadcaster_user_id': user_id } if not reward_id else { 'broadcaster_user_id': user_id, 'reward_id': reward_id }
    super().__init__(type = 'channel.channel_points_custom_reward.update', 
                     version =  1, 
                     condition =  condition, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class CustomRewardRemoveEvent(TwitchEvent):
  def __init__(self, user_id, session_id, reward_id = None):
    condition = { 'broadcaster_user_id': user_id } if not reward_id else { 'broadcaster_user_id': user_id, 'reward_id': reward_id }
    super().__init__(type = 'channel.channel_points_custom_reward.remove', 
                     version =  1, 
                     condition =  condition, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class CustomRewardRedemptionAddEvent(TwitchEvent):
  def __init__(self, user_id, session_id, reward_id = None):
    condition = { 'broadcaster_user_id': user_id } if not reward_id else { 'broadcaster_user_id': user_id, 'reward_id': reward_id }
    super().__init__(type = 'channel.channel_points_custom_reward_redemption.add', 
                     version =  1, 
                     condition =  condition, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class CustomRewardRedemptionUpdateEvent(TwitchEvent):
  def __init__(self, user_id, session_id, reward_id = None):
    condition = { 'broadcaster_user_id': user_id } if not reward_id else { 'broadcaster_user_id': user_id, 'reward_id': reward_id }
    super().__init__(type = 'channel.channel_points_custom_reward_redemption.update', 
                     version =  1, 
                     condition =  condition, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class StreamOnlineEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'stream.online', 
                     version = 1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })
    
class StreamOfflineEvent(TwitchEvent):
  def __init__(self, user_id, session_id):
    super().__init__(type = 'stream.offline', 
                     version =  1, 
                     condition =  { 'broadcaster_user_id': user_id }, 
                     transport =  { 'method': 'websocket', 'session_id': session_id })


import requests
import traceback
from ratelimit import limits, sleep_and_retry

class RateLimitedRequests:
  __CALLS_PER_MINUTE = 400
  __PERIOD = 60
  
  def __init__(self, cpm, period):
    self.__CALLS_PER_MINUTE = cpm
    self.__PERIOD = period
  
  @sleep_and_retry
  @limits(calls=__CALLS_PER_MINUTE, period=__PERIOD)
  def get(self, **params : dict) -> requests.Response:
    """Rate limited GET

    Returns:
        requests.Response: GET response
    """
    r = requests.get(**params)
    return r
  
    
  @sleep_and_retry
  @limits(calls=__CALLS_PER_MINUTE, period=__PERIOD)
  def post(self, **params : dict) -> requests.Response:
    """Rate limited POST

    Returns:
        requests.Response: POST response
    """
    r = requests.post(**params)
    return r


import os
import urllib
from .saferequests import *
    
class TwitchGQL_API:
  API_URL = "https://gql.twitch.tv/gql"
  CLIENT_ID = "kimne78kx3ncx6brgo4mv6wki5h1ko"
  DEFAULT_HEADERS = { "Client-ID": CLIENT_ID }
  REQ = RateLimitedRequests(400, 60)
    
  def get_clip(self, clip_id):
    """Get clip information from Twitch's GQL API.

    Args:
        clip_id (string): Clip ID

    Returns:
        dict: Clip info
    """
    content = [
      {
        "operationName": "VideoAccessToken_Clip",
        "variables": {
          "slug": f"{clip_id}"
        },
        "extensions": {
          "persistedQuery": {
            "version": 1,
            "sha256Hash": "36b89d2507fce29e5ca551df756d27c1cfe079e2609642b4390aa4c35796eb11"
          }
        }
      }
    ]
    r = self.REQ.post(url = self.API_URL, headers = self.DEFAULT_HEADERS, json = content)
    
    try:
      return r.json()[0]['data']['clip']
    except:
      return None
    
  def get_video(self, video_id):
    """Get video info.

    Args:
        video_id (string): Video ID

    Returns:
        dict: Video information
    """
    content = {
      "query": "query{video(id:\"" + video_id + "\"){title,thumbnailURLs(height:180,width:320),createdAt,lengthSeconds,owner{id,displayName}}}",
      "variables": {}
    }
    r = self.REQ.post(url = self.API_URL, headers = self.DEFAULT_HEADERS, json = content)
    
    try:
      return r.json()['data']['video']
    except:
      return None

  def download_clip(self, clip_id, filename, saveover = False):
    """Download a Twitch clip.

    Args:
        clip_id (string): Clip ID
        filename (string): Filename
        saveover (bool, optional): Overwrite existing files with the same name. Defaults to False.

    Returns:
        bool: Download succeeded/failed.
    """
    if (not saveover and os.path.exists(filename)):
      print(f"Skipping {clip_id} because the filename '{filename}' is taken.")
      return True
    
    info = self.get_clip(clip_id)
    
    url = info["videoQualities"][0]["sourceURL"]
    signature = info["playbackAccessToken"]["signature"]
    token = urllib.parse.quote(info["playbackAccessToken"]["value"])
    full_url = f"{url}?sig={signature}&token={token}"
    
    try:
      r = requests.get(full_url)
      with open(filename, 'wb') as outfile:
        outfile.write(r.content)
        return True
    except:
      return False
    
  def get_video_chapters(self, video_id):
    content = {
      "extensions" : { 
        "persistedQuery": {
          "sha256Hash": "8d2793384aac3773beab5e59bd5d6f585aedb923d292800119e03d40cd0f9b41",
          "version": 1
        }
      },
      "operationName": "VideoPlayer_ChapterSelectButtonVideo",
      "variables": {
        "videoID": video_id
      }
    }
    r = self.REQ.post(url = self.API_URL, headers = self.DEFAULT_HEADERS, json = content)
    try:
      return r.json()['data']['video']['moments']['edges']
    except:
      return []
    
  def __fetch_all_chat(self, video_id):
    """Fetches the chat messages without checking uniqueness or sorting.

    Args:
        video_id (string): Video ID

    Returns:
        list: List of chat messages
    """
    content = [
      {
        "operationName": "VideoCommentsByOffsetOrCursor",
        "variables": {
          "videoID": video_id,
          "contentOffsetSeconds": 0
        },
        "extensions": {
          "persistedQuery": {
            "version": 1,
            "sha256Hash": "b70a3591ff0f4e0313d126c6a1502d79a1c02baebb288227c582044aa76adf6a"
          }
        }
      }
    ]
    
    comments = []
    while True:
      r = self.REQ.post(url = self.API_URL, headers = self.DEFAULT_HEADERS, json = content)
      
      try:
        commentlist = r.json()[0]['data']['video']['comments']['edges']
      except KeyError:
        print("Improper response format.")
        return comments
      
      cursor = None
      for commentdata in commentlist:
        comment = commentdata['node']
        comments.append(comment)
        cursor = commentdata['cursor']
        
      try:
        has_next_page = r[0]['data']['video']['comments']['pageInfo']['hasNextPage']
      except KeyError:
        print("Issue with comment list. Missing pageInfo or hasNextPage.")
        return comments
      
      if has_next_page:
        if "contentOffsetSeconds" in content[0]['variables']:
          content[0]['variables'].pop("contentOffsetSeconds")
        content[0]['variables']['cursor'] = cursor
      else:
        return comments
      
  def get_chat_messages(self, video_id):
    """Get all chat messages and remove duplicates, sort by timestamp

    Args:
        video_id (string): Video ID

    Returns:
        list: List of chat messages
    """
    raw_chat = self.__fetch_all_chat(video_id)
    
    processed_chat = [i for n, i in enumerate(raw_chat) if i not in raw_chat[n + 1:]]
    
    processed_chat.sort(key=lambda a: a['contentOffsetSeconds'])
    return processed_chat
      